(ns wanijo.instance.routes
  (:require [compojure.core :refer [defroutes wrap-routes
                                    GET POST DELETE]]
            [ring.util.response :as resp]
            [formulare.core :as form]
            [wanijo.instance
             [view :as view]
             [domain :as domain]
             [forms :as forms-inst]]
            [wanijo.schema
             [domain :as domain-schema]
             [middleware :as middleware-schema]]
            [wanijo.framework.routing :refer [register! path]]
            [wanijo.attribute.domain :as domain-attr]))

(defn list! [schema-uuid req]
  (view/list! (domain-schema/find-by-uuid! schema-uuid)
              (domain/find-by-schema! schema-uuid)
              (forms-inst/with-attributes (domain-attr/required! schema-uuid))
              req))

(defn new! [req]
  (let [schema-uuid (get-in req [:params :schema-uuid])
        user-uuid (get-in req [:session :uuid])
        req-attrs (domain-attr/required! schema-uuid)
        form-def (forms-inst/with-attributes req-attrs)]
    (if (form/valid? form-def req)
      (let [form-data (form/form-data form-def req)
            req-attrs (domain-attr/required! schema-uuid)
            instance (forms-inst/form-data->instance form-data
                                                     req-attrs)]
        (domain/create! user-uuid
                        schema-uuid
                        instance)
        (resp/redirect (path :instance-list
                             (:params req))))
      (list! schema-uuid req))))

(defn instance! [uuid]
  (domain/full-instance-by-uuid! uuid))

(defn show! [uuid req]
  (let [user-uuid (-> req :session :uuid)
        instance (assoc (instance! uuid)
                        :starred
                        (domain/is-starred! uuid
                                            user-uuid))]
    (view/show! instance req)))

(defn form! [uuid attrs]
  (forms-inst/with-attributes attrs))

(defn edit-form! [uuid req]
  (let [instance (instance! uuid)
        attrs (domain-attr/find-by-instance! uuid)
        user-uuid (get-in req [:session :uuid])]
    (view/edit! instance
                (form! uuid attrs)
                (forms-inst/instance->form-data instance)
                (domain-schema/accessible-schemas! user-uuid)
                req)))

(defn edit! [uuid req]
  (let [attrs (domain-attr/find-by-instance! uuid)
        form-def (form! uuid attrs)
        instance (instance! uuid)]
    (if (form/valid? form-def req)
      (let [form-data (form/form-data form-def req)
            form-instance (forms-inst/form-data->instance form-data attrs)
            instance (assoc form-instance :uuid uuid)]
        (domain/edit! instance)
        (resp/redirect (path :instance-show instance)))
      (show! uuid req))))

(defn delete! [uuid req]
  (let [schema (domain-schema/find-by-instance! uuid)]
    (domain/delete! uuid)
    (resp/redirect (path :instance-list
                         {:schema-uuid (:uuid schema)}))))

(defn link-form! [schema-uuid]
  (forms-inst/link-form
   (domain/find-by-schema! schema-uuid)))

(defn link-selection! [uuid schema-uuid req]
  (view/link-selection (instance! uuid)
                       (domain-schema/find-by-uuid! schema-uuid)
                       (link-form! schema-uuid)
                       req))

(defn create-link! [uuid schema-uuid req]
  (let [form (link-form! schema-uuid)]
    (if (form/valid? form req)
      (let [form-data (form/form-data form req)]
        (domain/create-link! {:from uuid
                              :name (:name form-data)
                              :to (:instances form-data)
                              :by (get-in req [:session :uuid])})
        (resp/redirect (path :instance-edit-form {:uuid uuid})))
      (link-selection! uuid schema-uuid req))))

(defn delete-link! [uuid link-uuid req]
  (domain/delete-link! link-uuid)
  (resp/redirect (path :instance-edit-form {:uuid uuid})))

(defroutes routes
  (wrap-routes
   (GET (register! :instance-list "/instance/list/:schema-uuid")
        [schema-uuid :as req]
        (list! schema-uuid req))
   (let [schema-fn #(get-in % [:params :schema-uuid])]
     (middleware-schema/wrap-allowed-to-read schema-fn)))
  (POST (register! :instance-new "/instance/new") []
        new!)
  (GET (register! :instance-show "/instance/:uuid")
       [uuid :as req]
       (show! uuid req))
  (GET (register! :instance-edit-form "/instance/:uuid/edit")
       [uuid :as req]
       (edit-form! uuid req))
  (POST (register! :instance-edit "/instance/:uuid")
        [uuid :as req]
        (edit! uuid req))
  (DELETE (register! :instance-delete "/instance/:uuid")
          [uuid :as req]
          (delete! uuid req))
  (GET (register! :instance-link-selection
                  "/instance/:uuid/link/:schema-uuid")
       [uuid schema-uuid :as req]
       (link-selection! uuid schema-uuid req))
  (POST (register! :instance-link-create
                   "/instance/:uuid/link/:schema-uuid")
        [uuid schema-uuid :as req]
        (create-link! uuid schema-uuid req))
  (DELETE (register! :instance-link-delete
                     "/instance/:uuid/link/:link-uuid")
          [uuid link-uuid :as req]
          (delete-link! uuid link-uuid req)))
